/* FoundationAdditions.m created by lindberg on Mon 20-Dec-1999 */
/*-
 * Copyright (c) 2002-2006 Doug McClure
 *
 * Original code from eogenerator Copyright (c) 2001 Carl Lindberg and Mike Gentry
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#import "TGenAdditions.h"
#import <Foundation/Foundation.h>
#import <Foundation/NSPathUtilities.h>
#import <Foundation/NSDate.h>
#import <Foundation/NSScanner.h>
#import "BFEncodingUtilities.h"


@interface NSObject (KeyValueDeclarations)
- (id)valueForKeyPath:(NSString *)key;
@end

@interface NSObject (WarningAvoidance)
- (NSString *)name;  // method on EOAttribute/EOEntity etc.
@end


@implementation NSObject (ClassTestMethods)

- (BOOL)isString {
    return [self isKindOfClass:[NSString class]];
}

- (BOOL)isData {
    return [self isKindOfClass:[NSData class]];
}

- (BOOL)isInteger {
    if ( [self isKindOfClass:[NSString class]] ) {
        NSScanner *scanner = [NSScanner scannerWithString:(NSString *)self];
        double doubleValue;
        int intValue;

        if ( [scanner scanDouble:&doubleValue] && [scanner isAtEnd] ) {
            [scanner setScanLocation:0];

            if ( [scanner scanInt:&intValue] && [scanner isAtEnd]  ) {
                return (doubleValue == (double)intValue);
            }
        }
    }

    return NO;
}

- (BOOL)isReal {
    if ( [self isKindOfClass:[NSString class]] ) {
        NSScanner *scanner = [NSScanner scannerWithString:(NSString *)self];
        double doubleValue;
        int intValue;

        if ( [scanner scanDouble:&doubleValue] && [scanner isAtEnd] ) {
            [scanner setScanLocation:0];

            if ( [scanner scanInt:&intValue] && [scanner isAtEnd] ) {
                return (doubleValue != (double)intValue);
            }
        }
    }

    return NO;
}

- (BOOL)isArray {
    return [self isKindOfClass:[NSArray class]];
}

- (BOOL)isDictionary {
    return [self isKindOfClass:[NSDictionary class]];
}

@end


#ifdef COCOA

@implementation NSArray (MiscKeyValue)

- (id)valueForKey:(NSString *)key
{
    return [self valueForKeyPath:key];
}

- (id)valueForKeyPath:(NSString *)key
{
    if ([key isEqualToString:@"count"] || ![key hasPrefix:@"@"]) {// special case... are there others?
        return [super valueForKey:key];
    }
    else {
        NSString *myKey = [key substringFromIndex:1];
        NSString *restOfKey = @"";
        NSRange  dotRange = [myKey rangeOfString:@"."];
        char     selectorBuffer[128];
        SEL	 computeSelector;

        if (dotRange.length > 0) {
            restOfKey = [myKey substringFromIndex:NSMaxRange(dotRange)];
            myKey = [myKey substringToIndex:dotRange.location];
        }

        sprintf (selectorBuffer, "compute%sForKey:", [myKey cString]);
        if (islower(selectorBuffer[7])) selectorBuffer[7] = toupper(selectorBuffer[7]);
        computeSelector = sel_getUid (selectorBuffer);

        if (![self respondsToSelector:computeSelector]) {
            [NSException raise:NSInvalidArgumentException
                        format:@"-[%@ valueForKeyPath:]: NSArray does not implement selector %s required to compute aggregate %@",
                NSStringFromClass([self class]), selectorBuffer, key];
            return nil;
        }
        else {
            return [self performSelector:computeSelector withObject:restOfKey];
        }
    }
}

- (id)computeCountForKey:(NSString *)key
{
    return [NSNumber numberWithUnsignedInt:[self count]];
}

- (id)computeSumForKey:(NSString *)key
{
    NSArray *valueArray = [self valueForKey:key];
    int     i, count = [valueArray count];
    double  sum = 0.0;

    if (count == 0) return nil;

    for (i=0; i<count; i++) {
        sum += [[valueArray objectAtIndex:i] doubleValue];
    }

    return [NSNumber numberWithDouble:sum];
}

- (id)computeAvgForKey:(NSString *)key
{
    NSArray *valueArray = [self valueForKey:key];
    int     i, count = [valueArray count];
    double  sum = 0.0;

    if (count == 0) return nil;

    for (i=0; i<count; i++) {
        sum += [[valueArray objectAtIndex:i] doubleValue];
    }

    return [NSNumber numberWithDouble:(sum / (double)count)];
}

- (id)computeMaxForKey:(NSString *)key
{
    NSArray *valueArray = [self valueForKey:key];
    int     i, count = [valueArray count];
    double  max;

    if (count == 0) return nil;

    max = [[valueArray objectAtIndex:0] doubleValue];

    for (i=1; i<count; i++) {
        double currValue = [[valueArray objectAtIndex:i] doubleValue];
        if (currValue > max) max = currValue;
    }

    return [NSNumber numberWithDouble:max];
}

- (id)computeMinForKey:(NSString *)key
{
    NSArray *valueArray = [self valueForKey:key];
    int     i, count = [valueArray count];
    double  min;

    if (count == 0) return nil;

    min = [[valueArray objectAtIndex:0] doubleValue];

    for (i=1; i<count; i++) {
        double currValue = [[valueArray objectAtIndex:i] doubleValue];
        if (currValue < min) min = currValue;
    }

    return [NSNumber numberWithDouble:min];
}

@end
#endif


@implementation NSArray (TGen)

static int sortByNumber(id obj1, id obj2, void *context)
{
    NSString *path = (NSString *)context;

    double obj1val;
    double obj2val;

    if ( [path length] > 0 ) {
        obj1val = [[obj1 valueForKeyPath:path] doubleValue];
        obj2val = [[obj2 valueForKeyPath:path] doubleValue];
    }
    else {
        obj1val = [obj1 doubleValue];
        obj2val = [obj2 doubleValue];
    }

    if ( obj1val < obj2val )
        return -1;
    else if ( obj1val > obj2val )
        return 1;
    else
        return 0;
}

- (NSArray *)sortedNumberArray
{
    if ( [self count] > 1 )
        return self;
    else
        return [self sortedArrayUsingFunction:sortByNumber context:NULL];
}

- (id)computeSortedNumberArrayForKey:(NSString *)key
{
    NSString *restOfKey = @"";
    NSRange  dotRange = [key rangeOfString:@"."];

    if (dotRange.length > 0) {
        restOfKey = [key substringFromIndex:NSMaxRange(dotRange)];
        key = [key substringToIndex:dotRange.location];
    }

    if ( [restOfKey length] > 0 )
        return [[self sortedArrayUsingFunction:sortByNumber context:key] valueForKeyPath:restOfKey];
    else
        return [self sortedArrayUsingFunction:sortByNumber context:key];
}

@end


@implementation NSData (Base64Encoding)

/*"
 * Returns the Base-64 encoding string representing the receiver. Base-64
 * encoding (defined in RFC 1521) turns three bytes of binary data into four
 * bytes of ASCII data, and is used in MIME mail and other places. The resulting
 * string will not contain any newline characters; for more control use the
 * BFBase64EncodeData() function in BFEncodingUtilities.h.
"*/
- (NSString *)base64String
{
    return BFBase64EncodeData(self, nil);
}

@end


@implementation NSString (TGen)

/*"
 * Calls +#stringWithCStringNoCopy:length:freeWhenDone: with a value of YES for
 * freeWhenDone.
"*/
+ (id)stringWithCStringNoCopy:(char *)bytes length:(unsigned)length
{
    return [self stringWithCStringNoCopy:bytes length:length freeWhenDone:YES];
}

/*"
 * Returns an autoreleased string. Calls
 * -#initWithCStringNoCopy:length:freeWhenDone:.
"*/
+ (id)stringWithCStringNoCopy:(char *)bytes length:(unsigned)length freeWhenDone:(BOOL)freeBuffer
{
    return [[[self alloc] initWithCStringNoCopy:bytes length:length freeWhenDone:freeBuffer] autorelease];
}

@end

