/* TemplateGenerator.m created by doug on Mon 07-Jun-2000 */
/*-
 * Copyright (c) 2002-2006 Doug McClure
 *
 * Original code from eogenerator Copyright (c) 2001 Carl Lindberg and Mike Gentry
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#import "TemplateGenerator.h"
#import <MiscMerge.subproj/MiscMergeTemplate.h>
#import <MiscMerge.subproj/MiscMergeCommandBlock.h>
#import <MiscMerge.subproj/MiscMergeEngine.h>
#import <MiscFoundation.subproj/NSString+MiscAdditions.h>
#import "FoundationAdditions.h"
#import "SystemType.h"


@interface TemplateGenerator(Private)
- (void)processCommandLineArguments;
- (void)addDefaultSearchPaths;
@end


@implementation TemplateGenerator

- init
{
    [super init];

    verboseMode = NO;
    recursiveLookups = NO;
    
    userVariables = [[NSMutableDictionary alloc] init];

    delimStart = @"[$";
    delimEnd = @"$]";

    [self processCommandLineArguments];

    return self;
}

- (void)displayHelpAndExit:(int)exitStatus
{
    Printf(@"Usage: %@\n", [[NSProcessInfo processInfo] processName]);
    Printf(@"  Must specify either the template or input data to come from a file\n");
    Printf(@"%s",
           "Options:\n"
           "  -template <file>       Template file to use, can be in a search directory, - for stdin\n"
           "  -in <plistfile>        Plist file to read in to use as datasource, - for stdin\n"
           "  -out <file>            File to save output to, - for stdout\n"
           "\n"
           "  -delim <string>        Use <string> as start/end delimiters\n"
           "  -startdelim <string>   Use <string> as start delimiters\n"
           "  -enddelim <string>     Use <string> as end delimiters\n"
           "\n"
           "  -templatedir <path>    Directory to look for template files in\n"
           "  -define-<key> <value>  Add <key> with <value> to runtime variable list\n"
	   "  -recursive             Turn on recursive key-value lookup.\n"
           "  -force                 Force overwrite of read-only files\n\n"
           "  -verbose               Verbose output\n"
           "  -version               Print version number and exit\n"
           "  -help                  This help\n");

    exit(exitStatus);
}

- (int)argumentType:(NSString *)argument
{
    if ( [@"-version" hasPrefix:argument] )		return ArgumentVersion;
    else if ( [@"-help" hasPrefix:argument] )		return ArgumentHelp;
    else if ( [@"-verbose" hasPrefix:argument] )	return ArgumentVerbose;

    else if ( [@"-delim" hasPrefix:argument] )		return ArgumentDelim;
    else if ( [@"-startdelim" hasPrefix:argument] )	return ArgumentStartDelim;
    else if ( [@"-enddelim" hasPrefix:argument] )	return ArgumentEndDelim;

    else if ( [@"-recursive" hasPrefix:argument] )	return ArgumentRecursive;
    else if ( [@"-force" hasPrefix:argument] )		return ArgumentForce;

    else if ( [@"-template" hasPrefix:argument] )	return ArgumentTemplate;
    else if ( [@"-templatedir" hasPrefix:argument] )	return ArgumentTemplateDir;

    else if ( [@"-in" hasPrefix:argument] )		return ArgumentInputFile;
    else if ( [@"-out" hasPrefix:argument] )		return ArgumentOutputFile;

    else if ( [argument hasPrefix:@"-define-"] )	return ArgumentDefine;

    else if ( [argument hasPrefix:@"-"] )		return ArgumentInvalidOption;
    else return ArgumentUnknownOption;
}

- (void)processCommandLineArguments
{
    NSEnumerator *argEnum = [[[NSProcessInfo processInfo] arguments] objectEnumerator];
    NSString *currArg;

    [argEnum nextObject]; //skip argv[0]

    while ( currArg = [argEnum nextObject] ) {
        switch ( [self argumentType:currArg] ) {
            case ArgumentVersion:
                ErrPrintf(@"tgen Version 1.3");
                exit(0);
                break;

            case ArgumentHelp:
                [self displayHelpAndExit:0];
                break;

            case ArgumentVerbose:
                verboseMode = YES;
                break;

            case ArgumentDelim:
                if ( (currArg = [argEnum nextObject]) && [currArg length] > 0 ) {
                    delimStart = [currArg retain];
                    delimEnd = [currArg retain];
                }
                else
                    ErrPrintf(@"Warning: -delim option must specify a non-zero length string");
                break;

            case ArgumentStartDelim:
                if ( (currArg = [argEnum nextObject]) && [currArg length] > 0 )
                    delimStart = [currArg retain];
                else
                    ErrPrintf(@"Warning: -startdelim option must specify a non-zero length string");
                break;

            case ArgumentEndDelim:
                if ( (currArg = [argEnum nextObject]) && [currArg length] > 0 )
                    delimEnd = [currArg retain];
                else
                    ErrPrintf(@"Warning: -enddelim option must specify a non-zero length string");
                break;

            case ArgumentForce:
                forceOverwrites = YES;
                break;

	    case ArgumentRecursive:
	        recursiveLookups = YES;
		break;

            case ArgumentTemplateDir:
                if ( currArg = [argEnum nextObject] )
                    if ( ![[NSFileManager defaultManager] directoryExistsAtPath:currArg] )
                        ErrPrintf(@"Warning: Template search directory %@ does not exist", currArg);
                    else
                        [searchPaths addObject:currArg];
                else
                    ErrPrintf(@"Warning: template search directory must be a non-zero length path", currArg);
                break;

            case ArgumentTemplate:
                if ( templateFile != nil )
                    ErrPrintf(@"Warning: can only specify one template file, using the first one '%@'", templateFile);
                else if ( (currArg = [argEnum nextObject]) && [currArg length] > 0 )
                    templateFile = [currArg retain];
                else
                    ErrPrintf(@"Warning: template must be a non-zero length name");
                break;


            case ArgumentInputFile:
                if ( inputFile != nil )
                    ErrPrintf(@"Warning: can only specify one input file, using the first one '%@'", inputFile);
                else if ( (currArg = [argEnum nextObject]) && [currArg length] > 0 )
                    inputFile = [currArg retain];
                else
                    ErrPrintf(@"Warning: file must be a non-zero length name", currArg);
                break;

            case ArgumentOutputFile:
                if ( outputFile != nil )
                    ErrPrintf(@"Warning: can only specify one output file, using the first one '%@'", outputFile);
                else if ( (currArg = [argEnum nextObject]) && [currArg length] > 0 )
                    outputFile = [currArg retain];
                else
                    ErrPrintf(@"Warning: file must be a non-zero length name", currArg);
                break;


            case ArgumentDefine:
            {
                NSString *key = [currArg substringFromIndex:8];
                NSString *value = [argEnum nextObject];

                if ([key length] > 0 && value)
                    [userVariables setObject:value forKey:key];
            }
                break;


            case ArgumentUnknownOption:
            case ArgumentInvalidOption:
            default:
                ErrPrintf(@"Invalid option: '%@'", currArg);
                [self displayHelpAndExit:1];
                break;
        }
    }

    if ( templateFile == nil && inputFile == nil ) {
        ErrPrintf(@"Can not specify both the template and input to come from standard input.\nUse -help to see help.");
        exit(1);
    }

    [self addDefaultSearchPaths];
}

- (void)addDefaultSearchPaths
{
    NSArray *libraryPaths;
    int i;

#ifdef OPENSTEP_ONLY
    libraryPaths = NSStandardLibraryPaths();
#else
    libraryPaths = NSSearchPathForDirectoriesInDomains(
                        NSAllLibrariesDirectory,
                        NSUserDomainMask|NSLocalDomainMask|NSNetworkDomainMask,
                        YES);
#endif

    for (i=0; i<[libraryPaths count]; i++) {
        NSString *libPath = [libraryPaths objectAtIndex:i];
        [searchPaths addObject:[libPath stringByAppendingPathComponent:@"TGen"]];
    }

    [searchPaths addObject:[[NSBundle mainBundle] bundlePath]];

    if (verboseMode)
        ErrPrintf(@"Search path is:\n%@", searchPaths);
}



/*** GENERATION ***/

- (MiscMergeTemplate *)templateForName:(NSString *)name
{
    MiscMergeTemplate *template = nil;
    NSString *stdinput = nil;
    NSString *filename = nil;

    if ( name == nil || [name isEqualToString:@"-"] ) {
        NSFileHandle *stdinfile = [NSFileHandle fileHandleWithStandardInput];
        NSData *stdindata = [stdinfile readDataToEndOfFile];

        if ( [stdindata length] > 0 )
            stdinput = [[[NSString alloc] initWithData:stdindata encoding:NSUTF8StringEncoding] autorelease];
    }
    else {
        NSFileManager *fileManager = [NSFileManager defaultManager];

        if ([fileManager regularFileExistsAtPath:name])
            filename = name;
        else
            filename = [fileManager findFile:name inSearchPath:searchPaths];

        if (filename == nil) {
            ErrPrintf(@"Could not find template file '%@' in search path", name);
            exit(1);
        }
    }

    if ( filename != nil || stdinput != nil ) {
        if (verboseMode)
            ErrPrintf(@"Using template at path %@", filename);

        template = [[[MiscMergeTemplate alloc] init] autorelease];

        if ( verboseMode)
            ErrPrintf(@"Using delimiters '%@' and '%@'", delimStart, delimEnd);

        [template setStartDelimiter:delimStart endDelimiter:delimEnd];

        if ( filename != nil )
            [template parseContentsOfFile:filename];
        else if ( stdinput != nil )
            [template parseString:stdinput];
    }

    return template;
}


- (MiscMergeEngine *)engineForName:(NSString *)name
{
    MiscMergeEngine *engine = [[[MiscMergeEngine alloc] initWithTemplate:[self templateForName:name]] autorelease];
    [engine setUseRecursiveLookups:recursiveLookups];

    if ( [userVariables count] ) {
        NSEnumerator *userVariableEnum = [userVariables keyEnumerator];
        NSString *userKey;

        while (userKey = [userVariableEnum nextObject])
            [engine setEngineValue:[userVariables objectForKey:userKey] forKey:userKey];
    }

    return engine;
}


- (id)plistForName:(NSString *)name
{
    NSString *plistString = nil;
    id plist = nil;
    
    if ( name == nil || [name isEqualToString:@"-"] ) {
        NSFileHandle *stdinfile = [NSFileHandle fileHandleWithStandardInput];
        NSData *stdindata = [stdinfile readDataToEndOfFile];

        if ( [stdindata length] > 0 )
            plistString = [[[NSString alloc] initWithData:stdindata encoding:NSUTF8StringEncoding] autorelease];
        else
            plistString = @"{}";
    }
    else {
        plistString = [NSString stringWithContentsOfFile:name];
    }

    if ( plistString == nil ) {
        ErrPrintf(@"Error loading plist file '%@'", name);
        exit(1);
    }
    else {
        plist = [plistString propertyList];

        if ( plist == nil ) {
            ErrPrintf(@"File '%@' could not be parsed as a property list", name);
            exit(1);
        }
    }

    return plist;
}

- (void)generate
{
    NSFileManager *fileManager = [NSFileManager defaultManager];
    MiscMergeEngine *engine = [self engineForName:templateFile];
    NSString *plist = [self plistForName:inputFile];
    NSString *output;

    [engine setEngineValue:plist forKey:@"_topLevelObject"];
    output = [engine executeWithObject:plist sender:nil];

    if ( outputFile == nil || [outputFile isEqualToString:@"-"] ) {
        fwrite([output cString], 1, [output cStringLength], stdout);
    }
    else if ( !forceOverwrites &&
              [fileManager fileExistsAtPath:outputFile] &&
              ![fileManager isWritableFileAtPath:outputFile] ) {
        ErrPrintf(@"ERROR: Could not write file '%@'", outputFile);
    }
    else {
        NSData *fileData = [output dataUsingEncoding:[NSString defaultCStringEncoding]];

        if ( ![fileData writeToFile:outputFile atomically:YES] )
            ErrPrintf(@"ERROR: Could not write file '%@'", outputFile);
        else
            ErrPrintf(@"Generated %@\n", outputFile);
    }
}

@end
